#include "point.h"
#include "ray.h"
#include "model.h"
#include "vector.h"
#include "ModelOBJ.h"
#include <algorithm>
#include <iostream>
#include <string>
#include <limits>
#include <cmath>
#ifndef _bounds_h
#define _bounds_h
using namespace std;
/**
 * Implementation of Bound
 * This class provides an implementation of a Bound containing primitives of a 3D geometry.
 * @tparam T generic type.
 */
template<typename T> class Bounds{
    public:
        // Public attributes.
        /// The lowest point of this Bound.
        Point<T> pointMin;
        /// The upper point of this Bound.
        Point<T> pointMax;
        /// A boolean flag that represents if this Bound is struck by ray.
        bool visited;
        /// Array that contains the lowest and upper point of this Bound.
        Point3f bounds[2];
        /// FOR DEVELOP PURPOSES: 3D representation model.
        Model* model;
        /// FOR DEVELOP PURPOSES: 3D representation model.
        /// The face contained in this bound (Only if this bound is contained in a leaf node).
        int face = -2;
        /// A vector with the indexes of all faces in this bound.
        vector<int> faces;
        /**
         * Default constructor of the Bound class
         */
        Bounds(){
            pointMax = new Point<T>(0,0,0);
            pointMin = new Point<T>(0,0,0);
            visited = false;
            bounds[0] = pointMin;
            bounds[1] = pointMax;
        }

        /**
         * Constructor of the Bound class.
         * @param point1 one of the corners of this Bound.
         * @param point2 one of the corners of this Bound.
         * @param inter flag if this Bound is struck by ray.
         */

        Bounds(const Point<T> &point1, const Point<T> &point2, bool inter)
        : pointMin(min(point1.x,point2.x),min(point1.y,point2.y),min(point1.z,point2.z)),
        pointMax(max(point1.x,point2.x),max(point1.y,point2.y),max(point1.z,point2.z)){
            visited = inter;
            bounds[0] = pointMin;
            bounds[1] = pointMax;
        }

        /**
         * FOR DEVELOP PURPOSES:
         * Constructor of the Bound class.
         * @param point1 one of the corners of this Bound.
         * @param point2 one of the corners of this Bound.
         * @param model  3D representation of the model.
         */
        Bounds(const Point<T> &point1, const Point<T> &point2, Model* mod)
                : pointMin(min(point1.x,point2.x),min(point1.y,point2.y),min(point1.z,point2.z)),
                  pointMax(max(point1.x,point2.x),max(point1.y,point2.y),max(point1.z,point2.z)){
            model = mod;
            visited = false;
            bounds[0] = pointMin;
            bounds[1] = pointMax;
        }

        /**
         *
         * @param point1
         * @param point2
         * @param total_primitives
         */
        Bounds(const Point<T> &point1, const Point<T> &point2, int total_primitives)
                : pointMin(min(point1.x,point2.x),min(point1.y,point2.y),min(point1.z,point2.z)),
                  pointMax(max(point1.x,point2.x),max(point1.y,point2.y),max(point1.z,point2.z)){
            visited = false;
            bounds[0] = pointMin;
            bounds[1] = pointMax;
        }

        /**
         * Method to display the coords for the lowest point.
         */
        void write_lowest_point(){
            cout << "("+to_string(pointMin.x)+","+to_string(pointMin.y)+","
                +to_string(pointMin.z)+")" <<endl;
        }

        /**
         * Method to display the coords for the upper point.
         */
        void write_upper_point(){
            cout << "("+to_string(pointMax.x)+","+to_string(pointMax.y)+","
                +to_string(pointMax.z)+")" <<endl;
        }

        /**
         * Method to display the coords for the lowest and upper points.
         */
        void write_coords(){
            cout << "[Min:("+to_string(this->pointMin.x)+","+to_string(this->pointMin.y)+","
                +to_string(this->pointMin.z)+"),Max:("+to_string(this->pointMax.x)+","+to_string(this->pointMax.y)+","
                +to_string(this->pointMax.z)+")]" <<endl;
        }

        /**
         * Method to make an intersection test into the current bound.
         * @param test_ray ray to be tested
         * @return true if this bound is struck by ray, false otherwise.
         */
        bool intersection(const Ray3f &r){
            float txmin, txmax, tymin, tymax, tzmin, tzmax;

            bounds[0] = pointMin;
            bounds[1] = pointMax;
            Vector3f norm = r.direction.normalization();
            Vector3f invdir(1/norm.x,1/norm.y,1/norm.z);
            txmin = (bounds[r.sign[0]].x - r.origin.x) * invdir.x;
            txmax = (bounds[1-r.sign[0]].x - r.origin.x) * invdir.x; 
            tymin = (bounds[r.sign[1]].y - r.origin.y) * invdir.y; 
            tymax = (bounds[1-r.sign[1]].y - r.origin.y) * invdir.y;
            if ((txmin > tymax) || (tymin > txmax)) 
            {
                //cout << "1. Failed the intersection test at: " <<endl;
                //write_coords();
                return false;
            } 
            if (tymin > txmin) 
                txmin = tymin; 
            if (tymax < txmax) 
                txmax = tymax; 
            tzmin = (bounds[r.sign[2]].z - r.origin.z) * invdir.z; 
            tzmax = (bounds[1-r.sign[2]].z - r.origin.z) * invdir.z;
            if ((txmin > tzmax) || (tzmin > txmax)) 
            {
                //cout << "2. Failed the intersection test at: " <<endl;
                //write_coords();
                return false; 
            }
            if (tzmin > txmin) 
                txmin = tzmin; 
            if (tzmax < txmax) 
                txmax = tzmax;
            return true; 
        }

        /**
         * Method to make an intersection test with the current primitive.
         * Using Moller-Trumbore intersection algorithm.
         * @param test_ray ray to be tested
         * @return true if this primitive is struck by ray, false otherwise.
         */
        bool ray_triangle_intersect(const Ray3f &test_ray, const ModelOBJ::Mesh &mesh)
        {
            float tMax = numeric_limits<float>::infinity();
            const float EPSILON = 0.0000001;
            cout << "Current face: " << face << endl;
            Vector3f v1 = Vector3f(mesh.coordsV[mesh.indices[face]].vertex.x, mesh.coordsV[mesh.indices[face]].vertex.y,
                                   mesh.coordsV[mesh.indices[face]].vertex.z);
            cout << face << " x " << v1.x << " y " << v1.y << " z " << v1.z << endl;
            Vector3f v2 = Vector3f(mesh.coordsV[mesh.indices[face+1]].vertex.x, mesh.coordsV[mesh.indices[face+1]].vertex.y,
                                   mesh.coordsV[mesh.indices[face+1]].vertex.z);
            cout << face+1 << " x " << v2.x << " y " << v2.y << " z " << v2.z << endl;
            Vector3f v3 = Vector3f(mesh.coordsV[mesh.indices[face+2]].vertex.x, mesh.coordsV[mesh.indices[face+2]].vertex.y,
                                   mesh.coordsV[mesh.indices[face+2]].vertex.z);
            cout << face+2 << " x " << v3.x << " y " << v3.y << " z " << v3.z << endl;
            Vector3<T> edge1 = v2 - v1;
            Vector3<T> edge2 = v3 - v1;
            Vector3<T> cross_dir_edge2 = edge2.cross(test_ray.direction);
            float determinant = cross_dir_edge2.dot(edge1);
            if(abs(determinant) < EPSILON) return false;
            float inv_determinant = 1/determinant;
            Vector3<T> origin_vector = Vector3<T>(test_ray.origin.x, test_ray.origin.y, test_ray.origin.z);
            Vector3<T> ori_v1 = origin_vector - v1;
            float u = cross_dir_edge2.dot(ori_v1) * inv_determinant;
            if(u < 0 || u > 1) return false;
            Vector3<T> ori_v1_edge1 = edge1.cross(ori_v1);
            float v = ori_v1_edge1.dot(test_ray.direction) * inv_determinant;
            if(v < 0 || u + v > 1) return false;
            tMax = ori_v1_edge1.dot(edge2)* inv_determinant < tMax ? ori_v1_edge1.dot(edge2) * inv_determinant:tMax;
            //The point where the ray hit the plane is defined by the equation P = O + (tMax * R).
            Vector3<T> pHit = Vector3<T>(test_ray.origin.x + (tMax * test_ray.direction.x),
                                         test_ray.origin.y + (tMax * test_ray.direction.y),
                                         test_ray.origin.z + (tMax * test_ray.direction.z));
            cout << "The ray struck a primitive"<< endl;
            cout << "(" << pHit.x << ", " << pHit.y << ", " << pHit.z <<")"<< endl;
            return true;
        }

    bool ray_triangle_intersect_beta(const Ray3f &test_ray, const Vector3f &v1, const Vector3f &v2, const Vector3f &v3)
    {
        float tMax = numeric_limits<float>::infinity();
        const float EPSILON = 0.0000001;
        Vector3<T> edge1 = v2 - v1;
        Vector3<T> edge2 = v3 - v1;
        Vector3<T> cross_dir_edge2 = edge2.cross(test_ray.direction);
        float determinant = cross_dir_edge2.dot(edge1);
        if(abs(determinant) < EPSILON) return false;
        float inv_determinant = 1/determinant;
        Vector3<T> origin_vector = Vector3<T>(test_ray.origin.x, test_ray.origin.y, test_ray.origin.z);
        Vector3<T> ori_v1 = origin_vector - v1;
        float u = cross_dir_edge2.dot(ori_v1) * inv_determinant;
        if(u < 0 || u > 1) return false;
        Vector3<T> ori_v1_edge1 = edge1.cross(ori_v1);
        float v = ori_v1_edge1.dot(test_ray.direction) * inv_determinant;
        if(v < 0 || u + v > 1) return false;
        tMax = ori_v1_edge1.dot(edge2)* inv_determinant < tMax ? ori_v1_edge1.dot(edge2) * inv_determinant:tMax;
        //The point where the ray hit the plane is defined by the equation P = O + (tMax * R).
        Vector3<T> pHit = Vector3<T>(test_ray.origin.x + (tMax * test_ray.direction.x),
                                     test_ray.origin.y + (tMax * test_ray.direction.y),
                                     test_ray.origin.z + (tMax * test_ray.direction.z));
        //cout << "Current face: " << face << endl;
        //cout << "The ray struck a primitive"<< endl;
        //cout << "(" << pHit.x << ", " << pHit.y << ", " << pHit.z <<")"<< endl;
        return true;
    }

        /**
         * Method to calculate the maximum and minimum extent using the coords of all feces contained in this bound.
         * @param mesh the mesh of the current model.
         */
        void calculate_minimum_max_extent(const ModelOBJ::Mesh &mesh)
        {
            Point<T> new_min(numeric_limits<float>::max(),numeric_limits<float>::max(),numeric_limits<float>::max());
            Point<T> new_max(numeric_limits<float>::min(),numeric_limits<float>::min(),numeric_limits<float>::min());
            for(int f: faces)
            {
                const glm::vec3 &first_coord = mesh.coordsV[3*f].vertex;
                const glm::vec3 &second_coord = mesh.coordsV[(3*f)+1].vertex;
                const glm::vec3 &third_coord = mesh.coordsV[(3*f)+2].vertex;
                if(new_min.x > first_coord.x)
                    new_min.x = first_coord.x;
                if(new_min.y > first_coord.y)
                    new_min.y = first_coord.y;
                if(new_min.z > first_coord.z)
                    new_min.z = first_coord.z;
                if(new_min.x > second_coord.x)
                    new_min.x = second_coord.x;
                if(new_min.y > second_coord.y)
                    new_min.y = second_coord.y;
                if(new_min.z > second_coord.z)
                    new_min.z = second_coord.z;
                if(new_min.x > third_coord.x)
                    new_min.x = third_coord.x;
                if(new_min.y > third_coord.y)
                    new_min.y = third_coord.y;
                if(new_min.z > third_coord.z)
                    new_min.z = third_coord.z;
                if(new_max.x < first_coord.x)
                    new_max.x = first_coord.x;
                if(new_max.y < first_coord.y)
                    new_max.y = first_coord.y;
                if(new_max.z < first_coord.z)
                    new_max.z = first_coord.z;
                if(new_max.x < second_coord.x)
                    new_max.x = second_coord.x;
                if(new_max.y < second_coord.y)
                    new_max.y = second_coord.y;
                if(new_max.z < second_coord.z)
                    new_max.z = second_coord.z;
                if(new_max.x < third_coord.x)
                    new_max.x = third_coord.x;
                if(new_max.y < third_coord.y)
                    new_max.y = third_coord.y;
                if(new_max.z < third_coord.z)
                    new_max.z = third_coord.z;
            }
            pointMin = new_min;
            pointMax = new_max;
        }

        void print_faces()
        {
            for(int i = 0; i < faces.size(); i++)
                cout << faces[i] <<endl;
        }

        /**
         *
         * @param f_item
         * @return
         */
        bool contains(unsigned int f_item)
        {
            for(unsigned int item: faces)
            {
                if(f_item == item)
                    return true;
            }
            return false;
        }
};
#endif 