#include "bounds.h"
#include "model.h"
#include "utils.h"
#include <iostream>
#include <vector>
#include <cmath>
#include <queue>
#include <limits>
#include <cstdlib>
#include "ModelOBJ.h"
#include "model.h"
#include <algorithm>
#include <bits/stdc++.h>

using namespace std;
/**
 * Implementation of a BHV.
 *
 * This class provides an implementation of a BHV in C++.
 *
 * @tparam T generic type.
 */
template<typename T>class BVH{
    /**
     * Implementation of a Node.
     * Inner class that represents a Node of a binary tree.
     * @tparam U generic type
     */
    template<typename U> struct Node{
        public:
            // Public attributes.
            /// Bound that contains the scene.
            Bounds<U> content;
            /// Index that represents the i-th node of the tree.
            int index;
            /// Flag that represents if this Node is a leaf.
            bool leaf;
            /// Node that represents the 'father' of this node.
            struct Node<U>* father;
            /// Node that represents the 'left sibling' of this node.
            struct Node<U>* left;
            /// Index of the left node.
            int i_left;
            /// Node that represents the 'right sibling' of this node.
            struct Node<U>* right;
            /// Index of the right node.
            int i_right;
            /// Total of primitives in this node.
            unsigned int total_prim;
            /// Longest axis
            int axis;
            /**
             * Constructor of the Node class
             * @param cont Bound that contains the scene.
             */
            Node(const Bounds<U> &cont): content(cont){
                left = NULL;
                right = NULL;
                leaf = true;
                i_left = -2;
                i_right = -2;
                total_prim = 0;
                axis = 3;
            }

            Node(const Bounds<U> &cont, const unsigned int &t_prim): content(cont){
                left = NULL;
                right = NULL;
                leaf = true;
                i_left = -2;
                i_right = -2;
                total_prim = t_prim;
                axis = 3;
            }

            /**
             * Method to check if the current node has a 'left sibling'.
             * @return true if the current node has a 'left sibling', false otherwise.
             */
            bool has_left()
            {
                return left != NULL;
            }

            /**
             * Method to check if the current node has a 'right sibling'.
             * @return true if the current node has a 'right sibling', false otherwise.
             */
            bool has_right()
            {
                return right != NULL;
            }

            bool is_leaf()
            {
                return !has_right() && !has_left();
            }

    };
    public:
        // Public attributes for a BVH.
        /// Node representing the 'root' of the tree.
        struct Node<T>* root;
        /// Node representing the last node added.
        struct Node<T>* last_added;
        /// Current level of the tree.
        unsigned int current_level;
        /// Total nodes in the current level of the tree.
        unsigned int total_nodes;
        /// Vector containing all nodes.
        vector<Node<T>*> nodes;
        /// MESH
        ModelOBJ::Mesh mesh;

        /**
         * Constructor with parameters of the BHV class.
         * @param scene Bound that contains the scene.
         */
        BVH(Bounds<T> &scene){
            //root->father = NULL;
            root = new Node<T>(scene);
            root->index = 0;
            last_added = root;
            current_level = 1;
            nodes.push_back(root);
            cout << "Root added: DEBUG MODE" <<endl;
        }

        /**
         * Default constructor of the BVH class.
         */
        BVH(){
            root == NULL;
        }

        /**
         *
         * @param m
         */
        BVH(const ModelOBJ::Mesh &m):mesh(m){
            root = NULL;
        }

        /**
         * Method to return the root of the current tree.
         * @return the content of the root.
         */
        Bounds<T> get_root(){
            return root->get_content();
        }

        /**
         * Method to check if the current level of the tree is full.
         * @return true if the current level es full, false otherwise.
         */
        bool full_level(){
            return total_nodes == pow(2,current_level);
        }

        /**
         * Method to print the coord of the last node added.
         */
        void print_last_added(){
            last_added->get_content().write_upper_point();
        }

        void set_scene_extent(Point3f point_min, Point3f point_max, bool normalized)
        {
            if(normalized)
            {
                nodes[0]->content.pointMin = normalize(point_min, point_max, point_min);
                nodes[0]->content.pointMax = normalize(point_max, point_max, point_min);
            }else{
                nodes[0]->content.pointMin = point_min;
                nodes[0]->content.pointMax = point_max;
            }
        }

        /**
         * Method to auto construct a binary tree from a given scene.
         * @param scene Bound that contains the scene.
         * @param total_levels how many levels we can create of the given scene.
         * @param root_check only for debug purposes.
         */
        void auto_construct(const int &primitives_obj, bool root_check = false){
            queue<int> temp_bounds;
            temp_bounds.push(primitives_obj);
            total_nodes = 0;
            while (!temp_bounds.empty())
            {
                int current_no_primitives = temp_bounds.front();
                if(current_no_primitives != 1)
                {
                    string bin_rep = binary_representation(current_no_primitives);
                    int two_k = 0;
                    int r = 0;
                    if(bin_rep[1] == '1')
                    {
                        two_k = (int)pow(2,bin_rep.size()-1);
                        r = current_no_primitives - two_k;
                        temp_bounds.push(two_k);
                        temp_bounds.push(r);
                    }else
                    {
                        two_k = (int)pow(2,bin_rep.size()-1)/2;
                        r = (int) (current_no_primitives - pow(2,bin_rep.size()-1));
                        temp_bounds.push(two_k + r);
                        temp_bounds.push(two_k);
                    }
                    add_children(current_no_primitives);
                    temp_bounds.pop();
                }else
                {
                    add_children(current_no_primitives);
                    temp_bounds.pop();
                }
            }
        }

        /**
         * Method to add a Node in the tree.
         * @param scene Bound that contains the scene to be added.
         */
        void add_children(const int &t_primitives){
            Bounds<T> temp_bound(Point3f(0.0,0.0,0.0),
                                 Point3f(0.0,0.0,0.0), false);
            if(nodes.empty()){
                // if BVH is empty.
                root = new Node<T>(temp_bound, t_primitives);
                root->index = 0;
                last_added = root;
                current_level = 1;
                nodes.push_back(root);
            }else{
                // if is not empty.
                struct Node<T>* new_node;
                new_node = new Node<T>(temp_bound, t_primitives);
                float current_index = (float) last_added->index + 1;
                new_node->index = (int) current_index;
                last_added = new_node;
                int father_index = (int)floor((current_index-1.0)/2.0);
                // if current level is full
                if(full_level()){
                    current_level+=1;
                    total_nodes = 0;
                    Node<T>* cfather = nodes[father_index];
                    cfather->leaf = false;
                    cfather->left = new_node;
                    cfather->left->father = cfather;
                }else{
                    // if is not full, add this node to the current level.
                    Node<T>* cfather = nodes[father_index];
                    if((int)current_index == (2*father_index)+1){
                        cfather->leaf = false;
                        cfather->left = new_node;
                        cfather->left->father = cfather;
                    }else{
                        cfather->right = new_node;
                        cfather->right->father = cfather;
                    }
                }
                nodes.push_back(new_node);
                total_nodes +=1;
            }
        }

        /**
        * Method to fill up this BVH with the data of the current mesh.
        * @param current_centroids The centroids corresponding of the faces in this mesh.
        * @param normalized A flag indicating if this run is normalized
        */
        void fill_up_beta(vector<glm::vec3> current_centroids, bool normalized)
        {
            vector<Point3f> temp_cent;
            int current_face = 0;
            if(normalized)
            {
                for(glm::vec3 vector: current_centroids)
                {
                    Point3f temp_point(vector.x, vector.y, vector.z, 0, current_face);
                    temp_cent.emplace_back(normalize(temp_point, nodes[0]->content.pointMax, nodes[0]->content.pointMin));
                    current_face++;
                }
            }else{
                for(glm::vec3 vector: current_centroids)
                {
                    Point3f temp_point(vector.x, vector.y, vector.z, 0, current_face);
                    temp_cent.emplace_back(temp_point);
                    current_face++;
                }
            }
            int longest_axis = find_longest_axis(temp_cent);
            for(Point3f point: temp_cent)
            {
                point.l_axis = longest_axis;
            }
            sort(temp_cent.begin(), temp_cent.end(), sort_by_axis);
            for(Point3f point: temp_cent)
            {
                nodes[0]->content.faces.emplace_back(point.c_face);
            }
            for(int i = 1; i < nodes.size(); i++)
            {
                int father_index = (int)floor((i-1.0)/2.0);
                if((2*father_index)+1 == nodes[i]->index)
                {
                    //Si entra es que el indice actual es de un nodo izquierdo
                    for(int j = 0; j < nodes[i]->total_prim; j++)
                    {
                        nodes[i]->content.faces.emplace_back(nodes[father_index]->content.faces[j]);
                    }
                }else
                {
                    //Si entra a este else, es que el indice actual es de un nodo derecho
                    int start = (nodes[father_index]->total_prim - nodes[i]->total_prim);
                    for(int j = start; j < nodes[father_index]->total_prim; j++)
                    {
                        nodes[i]->content.faces.emplace_back(nodes[father_index]->content.faces[j]);
                    }
                }
                if(nodes[i]->is_leaf())
                {
                    nodes[i]->content.face = nodes[i]->content.faces[0];
                    nodes[i]->content.faces.clear();
                }else
                {
                    nodes[i]->content.calculate_minimum_max_extent(mesh);
                }
            }
        }

        void check_nodes()
        {
            for(Node<T>* n: nodes)
            {
                if(!n->is_leaf())
                {
                    if(n->content.faces.size() == n->total_prim)
                        continue;
                    else
                    {
                        cout << "Error: Primitivas conrrespondientes a este nodo no coinciden" << endl;
                        break;
                    }
                }else
                {
                    if(n->content.faces.size() == 0 && n->content.face != -2)
                        continue;
                    else
                    {
                        if(n->content.faces.size() == 1 && nodes.size() == 1)
                        {
                            continue;
                        }else
                        {
                            cout << "Error: Hoja corrupta, indice: " << n->index << endl;
                            cout << "Error: Faces tiene " << n->content.faces.size() << endl;
                            cout << "Error: Hoja es " << n->content.face << endl;
                            cout << "Total de nodos: " << nodes.size() << endl;
                            break;
                        }
                    }
                }
            }
        }

        /**
         * Method to make a preorder traversal in the current tree.
         * @param test_ray the ray to be tested in the intersection test.
         */
        bool traversal_v2(Ray3f &test_ray)
        {
            if(nodes[0]->content.faces.size() == 1)
            {
                nodes[0]->content.face = nodes[0]->content.faces[0];
                return test_r(test_ray, 0);
            }
            if(!nodes[0]->content.intersection(test_ray))
            {
                return false;
            }
            bool result = false;
            int current_total_nodes = nodes.size();
            int index = 0;
            while(index < current_total_nodes)
            {
                if(nodes[index]->content.faces.size() > 2)
                {
                    if(!nodes[(2*index)+1]->content.visited)
                    {
                        if(nodes[(2*index)+1]->content.intersection(test_ray))
                        {
                            index = (2*index)+1;
                            continue;
                        }
                    }
                    if(!nodes[(2*index)+2]->content.visited)
                    {
                        if(nodes[(2*index)+2]->content.intersection(test_ray))
                        {
                            index = (2*index)+2;
                            continue;
                        }
                    }
                    //Subes al padre anterior si es que en este nodo no hubo interseccion
                    if(index != 0)
                    {
                        nodes[index]->content.visited = true;
                        index = (int)floor((nodes[index]->index-1.0)/2.0);
                    }else
                    {
                        index = current_total_nodes;
                    }
                }else
                {
                    if(test_r(test_ray, (2*index)+1))
                    {
                        index = current_total_nodes;
                        result = true;
                    }else
                    {
                        if(nodes[index]->content.faces.size() == 2)
                        {
                            if(test_r(test_ray, (2*index)+2))
                            {
                                index = current_total_nodes;
                                result = true;
                            }else
                            {
                                //En este caso se tiene que regresar
                                if(index != 0)
                                {
                                    nodes[index]->content.visited = true;
                                    index = (int)floor((nodes[index]->index-1.0)/2.0);
                                }else
                                {
                                    index = current_total_nodes;
                                }
                            }
                        }
                    }
                }
            }
            return result;
        }

        void unmark_nodes()
        {
            for(Node<T>* node: nodes)
            {
                node->content.visited = false;
            }
        }

        /**
         * Method to do traversal in the vector of coords.
         * @param test_ray The centroids corresponding of the faces in this mesh.
         */
        bool linear_traversal_beta(Ray3f &test_ray)
        {
            bool struck = false;
            int we = 0;
            for(Node<T>* node: nodes)
            {
                if(node->leaf)
                {
                    int i= node->content.face;
                    Vector3f v1 = Vector3f(mesh.coordsV[3*i].vertex.x, mesh.coordsV[3*i].vertex.y, mesh.coordsV[3*i].vertex.z);
                    Vector3f v2 = Vector3f(mesh.coordsV[(3*i)+1].vertex.x, mesh.coordsV[(3*i)+1].vertex.y, mesh.coordsV[(3*i)+1].vertex.z);
                    Vector3f v3 = Vector3f(mesh.coordsV[(3*i)+2].vertex.x, mesh.coordsV[(3*i)+2].vertex.y, mesh.coordsV[(3*i)+2].vertex.z);
                    if(node->content.ray_triangle_intersect_beta(test_ray, v1, v2, v3))
                    {
                        struck = true;
                        break;
                    }
                }
                we++;
            }
            return struck;
        }

        bool test_r(Ray3f &test_ray, int element)
        {
            bool toca = false;
            int i= nodes[element]->content.face;
            Vector3f v1 = Vector3f(mesh.coordsV[3*i].vertex.x, mesh.coordsV[3*i].vertex.y, mesh.coordsV[3*i].vertex.z);
            Vector3f v2 = Vector3f(mesh.coordsV[(3*i)+1].vertex.x, mesh.coordsV[(3*i)+1].vertex.y, mesh.coordsV[(3*i)+1].vertex.z);
            Vector3f v3 = Vector3f(mesh.coordsV[(3*i)+2].vertex.x, mesh.coordsV[(3*i)+2].vertex.y, mesh.coordsV[(3*i)+2].vertex.z);
            if(nodes[element]->content.ray_triangle_intersect_beta(test_ray, v1, v2, v3))
            {
                toca = true;
            }
            return toca;
        }
};
