#include <assert.h>
#include <cmath>
#ifndef _vector3_h
#define _vector3_h
/**
 * Implementation of a 3D Vector.
 * This class provides an implementation of a 3D Vector in C++.
 * @tparam T generic type.
 */
template<typename T> class Vector3{
    public:
        /// The X direction coord of this Vector.
        T x;
        /// The Y direction coord of this Vector.
        T y;
        /// The Z direction coord of this Vector.
        T z;

        /**
         * Constructor of the Vector class.
         * @param x the X coord of this Vector.
         * @param y the Y coord of this Vector.
         * @param z the Z coord of this Vector.
         */
        Vector3(T x, T y, T z): x(x), y(y), z(z){}

        /**
         * Method to allow the access to coords like an array.
         * @param i the i-th space of the 'array'.
         * @return the corresponding coord of the i-th space.
         */
        T operator[](int i) const{
            assert(i >= 0 && i <= 2);
            if(i == 0) return x;
            if(i == 1) return y;
            return z;
        }

        /**
         * Method to make minus operation between two vectors.
         * @param vector the vector to be subtracted of the current vector.
         * @return the resulting vector of the operation.
         */
        inline Vector3<T> operator-(const Vector3<T> &vector)const{
            return Vector3<T>(x - vector.x, y - vector.y, z - vector.z);
        }

        /**
         * Method to make a add operation between to vectors.
         * @param vector the vector to be added to the current vector.
         * @return the resulting vector of the operation.
         */
        inline Vector3<T> operator+(const Vector3<T> &vector)const{
            return Vector3<T>(x + vector.x, y + vector.y, z + vector.z);
        }

        /**
         * Method to make a multiplication by scalar.
         * @param scalar the scalar that will multiply the vector.
         * @return the resulting vector of the operation.
         */
        inline Vector3<T> operator*(T scalar) const{
            return Vector3<T>(scalar * x, scalar * y, scalar * z);
        }


        /**
         * Method to make a cross product.
         * @param v the vector that
         * @return the resulting vector of the operation.
         */
        inline Vector3<T> cross(const Vector3<T> &v){
            return Vector3<T>((v.y * z) - (v.z * y), (v.z * x) - (v.x * z), (v.x * y) - (v.y * x));
        }

        /**
         * Method to make a dot product.
         * @param v1
         * @return the resulting vector of the operation.
         */
        inline T dot(const Vector3<T> &v){
            return v.x * x + v.y * y + v.z * z;
        }

        /**
         * Method to calculate the normal vector of this vector.
         * @return the corresponding normalized vector.
         */ 
        Vector3<T> normalization() const {
            float norm_fac = sqrt(pow(this->x,2) + pow(this->y,2) + pow(this->z,2));
            return Vector3<T>(this->x/norm_fac, this->y/norm_fac, this->z/norm_fac);
        }

};
/// Definition of the Vector3f (Vector3<float>) type.
typedef Vector3<float> Vector3f;
#endif 
